@@ -15,6 +15,7 @@
  *    except according to the terms contained in the LICENSE file.
  */
 
+#include <assert.h>
 #include <helpers.h>
 #include <ctype.h>
 #include <dirent.h>
@@ -39,119 +40,120 @@
 #  include <sys/sysctl.h>
 #endif
 
-#define RC_RUNLEVEL	RC_SVCDIR "/softlevel"
-
 #ifndef S_IXUGO
 #  define S_IXUGO (S_IXUSR | S_IXGRP | S_IXOTH)
 #endif
 
 /* File stream used for plugins to write environ vars to */
 FILE *rc_environ_fd = NULL;
+static int dirfds[RC_DIR_MAX];
 
-typedef struct rc_service_state_name {
-	RC_SERVICE state;
-	const char *name;
-} rc_service_state_name_t;
+int rc_dirfd(enum rc_dir dir) {
+	int flags = O_RDONLY | O_CLOEXEC | O_DIRECTORY;
+	int dirfd = AT_FDCWD;
+	const char *dirname;
 
-/* We MUST list the states below 0x10 first
- * The rest can be in any order */
-static const rc_service_state_name_t rc_service_state_names[] = {
-	{ RC_SERVICE_STARTED,     "started" },
-	{ RC_SERVICE_STOPPED,     "stopped" },
-	{ RC_SERVICE_STARTING,    "starting" },
-	{ RC_SERVICE_STOPPING,    "stopping" },
-	{ RC_SERVICE_INACTIVE,    "inactive" },
-	{ RC_SERVICE_WASINACTIVE, "wasinactive" },
-	{ RC_SERVICE_HOTPLUGGED,  "hotplugged" },
-	{ RC_SERVICE_FAILED,      "failed" },
-	{ RC_SERVICE_SCHEDULED,   "scheduled"},
-	{ RC_SERVICE_CRASHED,     "crashed"},
-	{ 0, NULL}
-};
+	if (dir >= RC_DIR_MAX)
+		return -1;
+	if (dirfds[dir] <= 0) {
+		switch (dir) {
+		case RC_DIR_INVALID:
+			return -1;
+		case RC_DIR_SYSCONF:
+			return dirfds[dir] = open(rc_sysconfdir(), flags);
+		case RC_DIR_USRCONF:
+			return dirfds[dir] = open(rc_usrconfdir(), flags);
+		case RC_DIR_RUNLEVEL:
+			return dirfds[dir] = open(rc_runleveldir(), flags);
+		case RC_DIR_SVCDIR:
+			dirname = rc_svcdir();
+			break;
+		default:
+			assert(dirnames[dir]);
+			dirname = dirnames[dir];
+			dirfd = rc_dirfd(RC_DIR_SVCDIR);
+			break;
+		}
+
+		if ((dirfds[dir] = openat(dirfd, dirname, flags)) == -1 && errno == ENOENT) {
+			if (mkdirat(dirfd, dirname, 0755) == -1 && errno != EEXIST)
+				return -1;
+			dirfds[dir] = openat(dirfd, dirname, flags);
+		}
+	}
+	return dirfds[dir];
+}
 
 #define LS_INITD	0x01
 #define LS_DIR		0x02
 static RC_STRINGLIST *
-ls_dir(const char *dir, int options)
+ls_dir(int targetfd, const char *dir, int options)
 {
 	DIR *dp;
 	struct dirent *d;
 	RC_STRINGLIST *list = NULL;
 	struct stat buf;
 	size_t l;
-	char file[PATH_MAX];
-	int r;
 
 	list = rc_stringlist_new();
-	if ((dp = opendir(dir)) == NULL)
-		return list;
-	while (((d = readdir(dp)) != NULL)) {
-		if (d->d_name[0] != '.') {
-			if (options & LS_INITD) {
-				/* Check that our file really exists.
-				 * This is important as a service maybe in a
-				 * runlevel, but could have been removed. */
-				snprintf(file, sizeof(file), "%s/%s",
-				    dir, d->d_name);
-				r = stat(file, &buf);
-				if (r != 0)
-					continue;
 
-				/* .sh files are not init scripts */
-				l = strlen(d->d_name);
-				if (l > 2 && d->d_name[l - 3] == '.' &&
-				    d->d_name[l - 2] == 's' &&
-				    d->d_name[l - 1] == 'h')
-					continue;
-			}
-			if (options & LS_DIR) {
-				snprintf(file, sizeof(file), "%s/%s",
-				    dir, d->d_name);
-				if (stat(file, &buf) != 0 ||
-				    !S_ISDIR(buf.st_mode))
-					continue;
-			}
-			rc_stringlist_add(list, d->d_name);
+	if (!(dp = do_opendirat(targetfd, dir)))
+		return list;
+
+	while (((d = readdir(dp)) != NULL)) {
+		if (d->d_name[0] == '.')
+			continue;
+		if (options & LS_INITD) {
+			/* Check that our file really exists.
+			 * This is important as a service maybe in a
+			 * runlevel, but could have been removed. */
+			if (fstatat(dirfd(dp), d->d_name, &buf, 0) != 0)
+				continue;
+
+			/* .sh files are not init scripts */
+			l = strlen(d->d_name);
+			if (l > 2 && d->d_name[l - 3] == '.' &&
+					d->d_name[l - 2] == 's' &&
+					d->d_name[l - 1] == 'h')
+				continue;
 		}
+		if (options & LS_DIR) {
+			if (fstatat(dirfd(dp), d->d_name, &buf, 0) != 0 || !S_ISDIR(buf.st_mode))
+				continue;
+		}
+		rc_stringlist_add(list, d->d_name);
 	}
 	closedir(dp);
 	return list;
 }
 
 static bool
-rm_dir(const char *pathname, bool top)
+rm_dir(int targetfd, const char *pathname, bool top)
 {
-	DIR *dp;
-	struct dirent *d;
-	char file[PATH_MAX];
-	struct stat s;
 	bool retval = true;
+	struct dirent *d;
+	struct stat s;
+	DIR *dp;
 
-	if ((dp = opendir(pathname)) == NULL)
+	if (!(dp = do_opendirat(targetfd, pathname)))
 		return false;
 
-	errno = 0;
-	while (((d = readdir(dp)) != NULL) && errno == 0) {
-		if (strcmp(d->d_name, ".") != 0 &&
-		    strcmp(d->d_name, "..") != 0)
-		{
-			snprintf(file, sizeof(file),
-			    "%s/%s", pathname, d->d_name);
-			if (stat(file, &s) != 0) {
+	while (!(d = readdir(dp))) {
+		if (strcmp(d->d_name, ".") == 0 || strcmp(d->d_name, "..") == 0)
+			continue;
+		if (fstatat(dirfd(dp), pathname, &s, 0) != 0) {
+			retval = false;
+			break;
+		}
+		if (S_ISDIR(s.st_mode)) {
+			if (!rm_dir(dirfd(dp), d->d_name, true)) {
 				retval = false;
 				break;
 			}
-			if (S_ISDIR(s.st_mode)) {
-				if (!rm_dir(file, true))
-				{
-					retval = false;
-					break;
-				}
-			} else {
-				if (unlink(file)) {
-					retval = false;
-					break;
-				}
+		} else {
+			if (unlinkat(dirfd(dp), d->d_name, 0) != 0) {
+				retval = false;
+				break;
 			}
 		}
 	}
@@ -160,7 +162,7 @@ rm_dir(const char *pathname, bool top)
 	if (!retval)
 		return false;
 
-	if (top && rmdir(pathname) != 0)
+	if (top && unlinkat(targetfd, pathname, AT_REMOVEDIR) != 0)
 		return false;
 
 	return true;
@@ -173,7 +175,8 @@ file_regex(const char *file, const char *regex)
 {
 	FILE *fp;
 	char *line = NULL;
-	size_t len = 0;
+	size_t size = 0;
+	ssize_t len = 0;
 	regex_t re;
 	bool retval = true;
 	int result;
@@ -190,7 +193,7 @@ file_regex(const char *file, const char *regex)
 		return false;
 	}
 
-	while ((rc_getline(&line, &len, fp))) {
+	while ((len = xgetline(&line, &size, fp)) != -1) {
 		char *str = line;
 		/* some /proc files have \0 separated content so we have to
 		   loop through the 'line' */
@@ -231,23 +234,6 @@ get_systype(void)
 	return systype;
 }
 
-static const char *
-detect_prefix(const char *systype)
-{
-#ifdef PREFIX
-	return RC_SYS_PREFIX;
-#else
-	if (systype) {
-		if (strcmp(systype, RC_SYS_NONE) == 0)
-			return NULL;
-		if (strcmp(systype, RC_SYS_PREFIX) == 0)
-			return RC_SYS_PREFIX;
-	}
-
-	return NULL;
-#endif
-}
-
 static const char *
 detect_container(const char *systype RC_UNUSED)
 {
@@ -257,6 +243,8 @@ detect_container(const char *systype RC_UNUSED)
 		       return NULL;
 		if (strcmp(systype, RC_SYS_JAIL) == 0)
 			return RC_SYS_JAIL;
+		if (strcmp(systype, RC_SYS_PODMAN) == 0)
+			return RC_SYS_PODMAN;
 	}
 
 	int jailed = 0;
@@ -265,6 +253,9 @@ detect_container(const char *systype RC_UNUSED)
 	if (sysctlbyname("security.jail.jailed", &jailed, &len, NULL, 0) == 0)
 		if (jailed == 1)
 			return RC_SYS_JAIL;
+
+	if (exists("/var/run/.containerenv"))
+		return RC_SYS_PODMAN;
 #endif
 
 #ifdef __linux__
@@ -285,6 +276,8 @@ detect_container(const char *systype RC_UNUSED)
 				return RC_SYS_SYSTEMD_NSPAWN;
 		if (strcmp(systype, RC_SYS_DOCKER) == 0)
 				return RC_SYS_DOCKER;
+		if (strcmp(systype, RC_SYS_PODMAN) == 0)
+			return RC_SYS_PODMAN;
 	}
 	if (file_regex("/proc/cpuinfo", "UML"))
 		return RC_SYS_UML;
@@ -302,6 +295,8 @@ detect_container(const char *systype RC_UNUSED)
 		return RC_SYS_RKT;
 	else if (file_regex("/proc/1/environ", "container=systemd-nspawn"))
 		return RC_SYS_SYSTEMD_NSPAWN;
+	else if (exists("/run/.containerenv"))
+		return RC_SYS_PODMAN;
 	else if (exists("/.dockerenv"))
 		return RC_SYS_DOCKER;
 	/* old test, I'm not sure when this was valid. */
@@ -356,12 +351,9 @@ rc_sys(void)
 	const char *sys;
 
 	systype = get_systype();
-	sys = detect_prefix(systype);
+	sys = detect_container(systype);
 	if (!sys) {
-		sys = detect_container(systype);
-		if (!sys) {
-			sys = detect_vm(systype);
-		}
+		sys = detect_vm(systype);
 	}
 
 	return sys;
@@ -379,13 +371,21 @@ rc_parse_service_state(RC_SERVICE state)
 	return NULL;
 }
 
+static int
+rc_parse_service_state_dirfd(RC_SERVICE state)
+{
+	for (size_t i = 0; rc_service_state_names[i].name; i++)
+		if (rc_service_state_names[i].state == state)
+			return rc_service_state_names[i].dir;
+	return -1;
+}
+
 /* Returns a list of all the chained runlevels used by the
  * specified runlevel in dependency order, including the
  * specified runlevel. */
 static void
 get_runlevel_chain(const char *runlevel, RC_STRINGLIST *level_list, RC_STRINGLIST *ancestor_list)
 {
-	char path[PATH_MAX];
 	RC_STRINGLIST *dirs;
 	RC_STRING *d, *parent;
 	const char *nextlevel;
@@ -410,8 +410,7 @@ get_runlevel_chain(const char *runlevel, RC_STRINGLIST *level_list, RC_STRINGLIS
 	 * We can now do exactly the above procedure for our chained
 	 * runlevels.
 	 */
-	snprintf(path, sizeof(path), "%s/%s", RC_RUNLEVELDIR, runlevel);
-	dirs = ls_dir(path, LS_DIR);
+	dirs = ls_dir(rc_dirfd(RC_DIR_RUNLEVEL), runlevel, LS_DIR);
 	TAILQ_FOREACH(d, dirs, entries) {
 		nextlevel = d->value;
 
@@ -438,18 +437,18 @@ get_runlevel_chain(const char *runlevel, RC_STRINGLIST *level_list, RC_STRINGLIS
 bool
 rc_runlevel_starting(void)
 {
-	return exists(RC_STARTING);
+	return existsat(rc_dirfd(RC_DIR_SVCDIR), "rc.starting");
 }
 
 bool
 rc_runlevel_stopping(void)
 {
-	return exists(RC_STOPPING);
+	return existsat(rc_dirfd(RC_DIR_SVCDIR), "rc.stopping");
 }
 
 RC_STRINGLIST *rc_runlevel_list(void)
 {
-	return ls_dir(RC_RUNLEVELDIR, LS_DIR);
+	return ls_dir(AT_FDCWD, rc_runleveldir(), LS_DIR);
 }
 
 char *
@@ -459,7 +458,7 @@ rc_runlevel_get(void)
 	char *runlevel = NULL;
 	size_t i;
 
-	if ((fp = fopen(RC_RUNLEVEL, "r"))) {
+	if ((fp = do_fopenat(rc_dirfd(RC_DIR_SVCDIR), "softlevel", O_RDONLY))) {
 		runlevel = xmalloc(sizeof(char) * PATH_MAX);
 		if (fgets(runlevel, PATH_MAX, fp)) {
 			i = strlen(runlevel) - 1;
@@ -481,11 +480,12 @@ rc_runlevel_get(void)
 bool
 rc_runlevel_set(const char *runlevel)
 {
-	FILE *fp = fopen(RC_RUNLEVEL, "w");
+	FILE *fp;
 
-	if (!fp)
+	if (!(fp = do_fopenat(rc_dirfd(RC_DIR_SVCDIR), "softlevel", O_WRONLY | O_CREAT | O_TRUNC)))
 		return false;
-	fprintf(fp, "%s", runlevel);
+
+	fputs(runlevel, fp);
 	fclose(fp);
 	return true;
 }
@@ -493,14 +493,11 @@ rc_runlevel_set(const char *runlevel)
 bool
 rc_runlevel_exists(const char *runlevel)
 {
-	char path[PATH_MAX];
 	struct stat buf;
 
-	if (!runlevel || strcmp(runlevel, "") == 0 || strcmp(runlevel, ".") == 0 ||
-		strcmp(runlevel, "..") == 0)
+	if (!runlevel || strcmp(runlevel, "") == 0 || strcmp(runlevel, ".") == 0 || strcmp(runlevel, "..") == 0)
 		return false;
-	snprintf(path, sizeof(path), "%s/%s", RC_RUNLEVELDIR, runlevel);
-	if (stat(path, &buf) == 0 && S_ISDIR(buf.st_mode))
+	if (fstatat(rc_dirfd(RC_DIR_RUNLEVEL), runlevel, &buf, 0) == 0 && S_ISDIR(buf.st_mode))
 		return true;
 	return false;
 }
@@ -508,22 +505,29 @@ rc_runlevel_exists(const char *runlevel)
 bool
 rc_runlevel_stack(const char *dst, const char *src)
 {
-	char d[PATH_MAX], s[PATH_MAX];
+	char *d, *s;
+	int r;
 
 	if (!rc_runlevel_exists(dst) || !rc_runlevel_exists(src))
 		return false;
-	snprintf(s, sizeof(s), "../%s", src);
-	snprintf(d, sizeof(s), "%s/%s/%s", RC_RUNLEVELDIR, dst, src);
-	return (symlink(s, d) == 0 ? true : false);
+	xasprintf(&s, "../%s", src);
+	xasprintf(&d, "%s/%s/%s", rc_runleveldir(), dst, src);
+	r = symlink(s, d);
+	free(d);
+	free(s);
+	return (r == 0);
 }
 
 bool
 rc_runlevel_unstack(const char *dst, const char *src)
 {
-	char path[PATH_MAX];
+	char *path;
+	int r;
 
-	snprintf(path, sizeof(path), "%s/%s/%s", RC_RUNLEVELDIR, dst, src);
-	return (unlink(path) == 0 ? true : false);
+	xasprintf(&path, "%s/%s/%s", rc_runleveldir(), dst, src);
+	r = unlink(path);
+	free(path);
+	return (r == 0);
 }
 
 RC_STRINGLIST *
@@ -538,13 +542,199 @@ rc_runlevel_stacks(const char *runlevel)
 	return stack;
 }
 
+enum scriptdirs_entries {
+	SCRIPTDIR_USR,
+#ifdef RC_LOCAL_PREFIX
+	SCRIPTDIR_LOCAL,
+#endif
+	SCRIPTDIR_SYS,
+#ifdef RC_PKG_PREFIX
+	SCRIPTDIR_PKG,
+#endif
+	SCRIPTDIR_SVC,
+	SCRIPTDIR_MAX,
+	SCRIPTDIR_CAP
+};
+
+static const char * const scriptdirs[SCRIPTDIR_CAP] = {
+#ifdef RC_LOCAL_PREFIX
+	RC_LOCAL_PREFIX "/etc",
+#endif
+	RC_SYSCONFDIR,
+#ifdef RC_PKG_PREFIX
+	RC_PKG_PREFIX "/etc",
+#endif
+	RC_SVCDIR
+};
+
+static struct {
+	bool set;
+	char *svcdir;
+	char *usrconfdir;
+	char *runleveldir;
+	const char *scriptdirs[ARRAY_SIZE(scriptdirs)];
+} rc_dirs = {
+	.scriptdirs = {
+#ifdef RC_LOCAL_PREFIX
+		[SCRIPTDIR_LOCAL] = RC_LOCAL_PREFIX "/etc/user",
+#endif
+		[SCRIPTDIR_SYS] = RC_SYSCONFDIR "/user",
+#ifdef RC_PKG_PREFIX
+		[SCRIPTDIR_PKG] = RC_PKG_PREFIX "/etc/user",
+#endif
+	}
+};
+
+static void
+free_rc_dirs(void)
+{
+	free(rc_dirs.runleveldir);
+	rc_dirs.runleveldir = NULL;
+	free(rc_dirs.svcdir);
+	rc_dirs.svcdir = NULL;
+	rc_dirs.scriptdirs[0] = NULL;
+}
+
+static bool is_user = false;
+
+bool
+rc_is_user(void)
+{
+	return is_user;
+}
+
+void
+rc_set_user(void)
+{
+	char *env;
+	if (is_user)
+		return;
+
+	is_user = true;
+	rc_dirs.set = true;
+	setenv("RC_USER_SERVICES", "yes", true);
+
+	if ((env = getenv("XDG_CONFIG_HOME"))) {
+		xasprintf(&rc_dirs.usrconfdir, "%s/rc", env);
+	} else if ((env = getenv("HOME"))) {
+		xasprintf(&rc_dirs.usrconfdir, "%s/.config/rc", env);
+	} else {
+		fprintf(stderr, "XDG_CONFIG_HOME and HOME unset.\n");
+		exit(EXIT_FAILURE);
+	}
+
+	xasprintf(&rc_dirs.runleveldir, "%s/runlevels", rc_dirs.usrconfdir);
+
+	if (!(env = getenv("XDG_RUNTIME_DIR"))) {
+		/* FIXME: fallback to something else? */
+		fprintf(stderr, "XDG_RUNTIME_DIR unset.\n");
+		exit(EXIT_FAILURE);
+	}
+
+	xasprintf(&rc_dirs.svcdir, "%s/openrc", env);
+	atexit(free_rc_dirs);
+
+	rc_dirs.scriptdirs[SCRIPTDIR_USR] = rc_dirs.usrconfdir;
+	rc_dirs.scriptdirs[SCRIPTDIR_SVC] = rc_dirs.svcdir;
+
+	for (size_t i = 0; i < RC_DIR_MAX; i++) {
+		if (dirfds[i] == -1)
+			continue;
+		close(dirfds[i]);
+		dirfds[i] = -1;
+	}
+}
+
+const char * const *
+rc_scriptdirs(void)
+{
+	if (rc_dirs.set)
+		return rc_dirs.scriptdirs;
+	return scriptdirs;
+}
+
+size_t
+rc_scriptdirfds(const int **fds)
+{
+	static int scriptdirfds[SCRIPTDIR_MAX];
+	static size_t len = 0;
+
+	if (len == 0) {
+		const char * const *current_scriptdirs = rc_scriptdirs();
+		for (size_t i = 0; current_scriptdirs[i]; i++) {
+			int fd = open(current_scriptdirs[i], O_RDONLY | O_DIRECTORY | O_CLOEXEC);
+
+			if (fd == -1)
+				continue;
+
+			scriptdirfds[len++] = fd;
+		}
+	}
+
+	*fds = scriptdirfds;
+	return len;
+}
+
+const char *
+rc_sysconfdir(void)
+{
+	return RC_SYSCONFDIR;
+}
+
+const char *
+rc_usrconfdir(void)
+{
+	if (rc_dirs.set)
+		return rc_dirs.usrconfdir;
+
+	return NULL;
+}
+
+const char *
+rc_runleveldir(void)
+{
+	if (rc_dirs.set)
+		return rc_dirs.runleveldir;
+	return RC_RUNLEVELDIR;
+}
+
+const char *
+rc_svcdir(void)
+{
+	const char *path = rc_dirs.set ? rc_dirs.svcdir : RC_SVCDIR;
+	mkdir(path, 0755);
+	return path;
+}
+
+static ssize_t
+safe_readlink(int targetfd, const char *path, char **buf, size_t bufsiz)
+{
+	char *buffer;
+	ssize_t r;
+
+	for (;; bufsiz += PATH_MAX) {
+		buffer = xmalloc(bufsiz);
+		r = readlinkat(targetfd, path, buffer, bufsiz);
+		if (r < 0) {
+			free(buffer);
+			return -errno;
+		}
+		if ((size_t)r < bufsiz) {
+			buffer[r] = '\0';
+			*buf = buffer;
+			return r;
+		}
+		free(buffer);
+		if (bufsiz >= SSIZE_MAX - PATH_MAX)
+			return -ENOMEM;
+	}
+}
+
 /* Resolve a service name to its full path */
 char *
 rc_service_resolve(const char *service)
 {
-	char buffer[PATH_MAX];
-	char file[PATH_MAX];
-	int r;
+	char *buffer;
 	struct stat buf;
 
 	if (!service)
@@ -554,40 +744,21 @@ rc_service_resolve(const char *service)
 		return xstrdup(service);
 
 	/* First check started services */
-	snprintf(file, sizeof(file), RC_SVCDIR "/%s/%s", "started", service);
-	if (lstat(file, &buf) || !S_ISLNK(buf.st_mode)) {
-		snprintf(file, sizeof(file), RC_SVCDIR "/%s/%s",
-		    "inactive", service);
-		if (lstat(file, &buf) || !S_ISLNK(buf.st_mode))
-			*file = '\0';
+	if (fstatat(rc_dirfd(RC_DIR_STARTED), service, &buf, AT_SYMLINK_NOFOLLOW) == 0 && S_ISLNK(buf.st_mode))
+		if (safe_readlink(rc_dirfd(RC_DIR_STARTED), service, &buffer, buf.st_size + 1) >= 0)
+			return buffer;
+	if (fstatat(rc_dirfd(RC_DIR_INACTIVE), service, &buf, AT_SYMLINK_NOFOLLOW) == 0 && S_ISLNK(buf.st_mode))
+		if (safe_readlink(rc_dirfd(RC_DIR_INACTIVE), service, &buffer, buf.st_size + 1) >= 0)
+			return buffer;
+
+	for (const char * const *dirs = rc_scriptdirs(); *dirs; dirs++) {
+		char *file = NULL;
+		xasprintf(&file, "%s/init.d/%s", *dirs, service);
+		if (stat(file, &buf) == 0)
+			return file;
+		free(file);
 	}
 
-	if (*file) {
-		memset(buffer, 0, sizeof(buffer));
-		r = readlink(file, buffer, sizeof(buffer)-1);
-		if (r > 0)
-			return xstrdup(buffer);
-	}
-
-#ifdef RC_LOCAL_INITDIR
-	/* Nope, so lets see if the user has written it */
-	snprintf(file, sizeof(file), RC_LOCAL_INITDIR "/%s", service);
-	if (stat(file, &buf) == 0)
-		return xstrdup(file);
-#endif
-
-	/* System scripts take precedence over 3rd party ones */
-	snprintf(file, sizeof(file), RC_INITDIR "/%s", service);
-	if (stat(file, &buf) == 0)
-		return xstrdup(file);
-
-#ifdef RC_PKG_INITDIR
-	/* Check RC_PKG_INITDIR */
-	snprintf(file, sizeof(file), RC_PKG_INITDIR "/%s", service);
-	if (stat(file, &buf) == 0)
-		return xstrdup(file);
-#endif
-
 	return NULL;
 }
 
@@ -608,8 +779,8 @@ rc_service_exists(const char *service)
 
 	/* .sh files are not init scripts */
 	if (len > 2 && service[len - 3] == '.' &&
-	    service[len - 2] == 's' &&
-	    service[len - 1] == 'h') {
+			service[len - 2] == 's' &&
+			service[len - 1] == 'h') {
 		errno = EINVAL;
 		return false;
 	}
@@ -653,19 +824,22 @@ rc_service_extra_commands(const char *service)
 	snprintf(cmd, l, OPTSTR, svc);
 	free(svc);
 
-	if ((fp = popen(cmd, "r"))) {
-		rc_getline(&buffer, &len, fp);
+	if (!(fp = popen(cmd, "r"))) {
+		free(cmd);
+		return NULL;
+	}
+
+	if (xgetline(&buffer, &len, fp) != -1) {
 		p = buffer;
 		commands = rc_stringlist_new();
 
 		while ((token = strsep(&p, " ")))
 			if (token[0] != '\0')
 				rc_stringlist_add(commands, token);
-
-		pclose(fp);
-		free(buffer);
 	}
 
+	pclose(fp);
+	free(buffer);
 	free(cmd);
 	return commands;
 }
@@ -677,7 +851,7 @@ rc_service_description(const char *service, const char *option)
 	char *svc;
 	char *cmd;
 	char *desc = NULL;
-	size_t len = 0;
+	size_t size = 0;
 	FILE *fp;
 	size_t l;
 
@@ -692,7 +866,10 @@ rc_service_description(const char *service, const char *option)
 	snprintf(cmd, l, DESCSTR, svc, *option ? "_" : "", option);
 	free(svc);
 	if ((fp = popen(cmd, "r"))) {
-		rc_getline(&desc, &len, fp);
+		if (xgetline(&desc, &size, fp) == -1) {
+			free(desc);
+			desc = NULL;
+		}
 		pclose(fp);
 	}
 	free(cmd);
@@ -702,26 +879,22 @@ rc_service_description(const char *service, const char *option)
 bool
 rc_service_in_runlevel(const char *service, const char *runlevel)
 {
-	char file[PATH_MAX];
+	char *file;
+	bool r;
 
-	snprintf(file, sizeof(file), RC_RUNLEVELDIR "/%s/%s",
-	    runlevel, basename_c(service));
-	return exists(file);
+	xasprintf(&file, "%s/%s/%s", rc_runleveldir(), runlevel, basename_c(service));
+	r = exists(file);
+	free(file);
+	return r;
 }
 
 bool
 rc_service_mark(const char *service, const RC_SERVICE state)
 {
-	char file[PATH_MAX];
 	int i = 0;
-	int skip_state = -1;
 	const char *base;
 	char *init = rc_service_resolve(service);
 	bool skip_wasinactive = false;
-	int s;
-	char was[PATH_MAX];
-	RC_STRINGLIST *dirs;
-	RC_STRING *dir;
 	int serrno;
 
 	if (!init)
@@ -729,21 +902,17 @@ rc_service_mark(const char *service, const RC_SERVICE state)
 
 	base = basename_c(service);
 	if (state != RC_SERVICE_STOPPED) {
+		int state_dirfd = rc_dirfd(rc_parse_service_state_dirfd(state));
 		if (!exists(init)) {
 			free(init);
 			return false;
 		}
 
-		snprintf(file, sizeof(file), RC_SVCDIR "/%s/%s",
-		    rc_parse_service_state(state), base);
-		if (exists(file))
-			unlink(file);
-		i = symlink(init, file);
-		if (i != 0) {
+		unlinkat(state_dirfd, base, 0);
+		if ((symlinkat(init, state_dirfd, base)) != 0) {
 			free(init);
 			return false;
 		}
-		skip_state = state;
 	}
 
 	if (state == RC_SERVICE_HOTPLUGGED || state == RC_SERVICE_FAILED) {
@@ -753,78 +922,68 @@ rc_service_mark(const char *service, const RC_SERVICE state)
 
 	/* Remove any old states now */
 	for (i = 0; rc_service_state_names[i].name; i++) {
-		s = rc_service_state_names[i].state;
+		int state_dirfd = rc_dirfd(rc_service_state_names[i].dir);
+		RC_SERVICE s = rc_service_state_names[i].state;
 
-		if ((s != skip_state &&
-			s != RC_SERVICE_STOPPED &&
-			s != RC_SERVICE_HOTPLUGGED &&
-			s != RC_SERVICE_SCHEDULED) &&
-		    (!skip_wasinactive || s != RC_SERVICE_WASINACTIVE))
-		{
-			snprintf(file, sizeof(file), RC_SVCDIR "/%s/%s",
-			    rc_service_state_names[i].name, base);
-			if (exists(file)) {
-				if ((state == RC_SERVICE_STARTING ||
-					state == RC_SERVICE_STOPPING) &&
-				    s == RC_SERVICE_INACTIVE)
-				{
-					snprintf(was, sizeof(was),
-					    RC_SVCDIR "/%s/%s",
-					    rc_parse_service_state(RC_SERVICE_WASINACTIVE),
-					    base);
-					if (symlink(init, was) == -1) {
-						free(init);
-						return false;
-					}
-					skip_wasinactive = true;
-				}
-				if (unlink(file) == -1) {
+		if (s == state)
+			continue;
+
+		switch (s) {
+		case RC_SERVICE_STOPPED:
+		case RC_SERVICE_HOTPLUGGED:
+		case RC_SERVICE_SCHEDULED:
+			continue;
+		case RC_SERVICE_WASINACTIVE:
+			if (skip_wasinactive)
+				continue;
+			/* fall-through */
+		default:
+			if (!existsat(state_dirfd, base))
+				continue;
+			if ((state == RC_SERVICE_STARTING || state == RC_SERVICE_STOPPING) && s == RC_SERVICE_INACTIVE) {
+				if (symlinkat(init, rc_dirfd(RC_DIR_WASINACTIVE), base) == -1) {
 					free(init);
 					return false;
 				}
+				skip_wasinactive = true;
+			}
+			if (unlinkat(state_dirfd, base, 0) == -1) {
+				free(init);
+				return false;
 			}
 		}
 	}
 
 	/* Remove the exclusive state if we're inactive */
-	if (state == RC_SERVICE_STARTED ||
-	    state == RC_SERVICE_STOPPED ||
-	    state == RC_SERVICE_INACTIVE)
-	{
-		snprintf(file, sizeof(file), RC_SVCDIR "/%s/%s",
-		    "exclusive", base);
-		unlink(file);
-	}
+	if (state == RC_SERVICE_STARTED || state == RC_SERVICE_STOPPED || state == RC_SERVICE_INACTIVE)
+		unlinkat(rc_dirfd(RC_DIR_EXCLUSIVE), base, 0);
 
 	/* Remove any options and daemons the service may have stored */
 	if (state == RC_SERVICE_STOPPED) {
-		snprintf(file, sizeof(file), RC_SVCDIR "/%s/%s",
-		    "options", base);
-		rm_dir(file, true);
-
-		snprintf(file, sizeof(file), RC_SVCDIR "/%s/%s",
-		    "daemons", base);
-		rm_dir(file, true);
-
+		rm_dir(rc_dirfd(RC_DIR_OPTIONS), base, true);
+		rm_dir(rc_dirfd(RC_DIR_DAEMONS), base, true);
 		rc_service_schedule_clear(service);
 	}
 
 	/* These are final states, so remove us from scheduled */
 	if (state == RC_SERVICE_STARTED || state == RC_SERVICE_STOPPED) {
-		snprintf(file, sizeof(file), RC_SVCDIR "/%s", "scheduled");
-		dirs = ls_dir(file, 0);
-		TAILQ_FOREACH(dir, dirs, entries) {
-			snprintf(was, sizeof(was), "%s/%s/%s",
-			    file, dir->value, base);
-			unlink(was);
+		DIR *dp = do_dopendir(rc_dirfd(RC_DIR_SCHEDULED));
+		if (dp) {
+			struct dirent *d;
+			while ((d = readdir(dp))) {
+				char *was;
 
-			/* Try and remove the dir; we don't care about errors */
-			snprintf(was, sizeof(was), "%s/%s", file, dir->value);
-			serrno = errno;
-			rmdir(was);
-			errno = serrno;
+				xasprintf(&was, "%s/%s", d->d_name, base);
+				unlinkat(dirfd(dp), was, 0);
+				free(was);
+
+				serrno = errno;
+				unlinkat(dirfd(dp), d->d_name, AT_REMOVEDIR);
+				errno = serrno;
+			}
+
+			closedir(dp);
 		}
-		rc_stringlist_free(dirs);
 	}
 	free(init);
 	return true;
@@ -833,17 +992,12 @@ rc_service_mark(const char *service, const RC_SERVICE state)
 RC_SERVICE
 rc_service_state(const char *service)
 {
-	int i;
-	int state = RC_SERVICE_STOPPED;
-	char file[PATH_MAX];
-	RC_STRINGLIST *dirs;
-	RC_STRING *dir;
 	const char *base = basename_c(service);
+	int state = RC_SERVICE_STOPPED;
+	int i;
 
 	for (i = 0; rc_service_state_names[i].name; i++) {
-		snprintf(file, sizeof(file), RC_SVCDIR "/%s/%s",
-		    rc_service_state_names[i].name, base);
-		if (exists(file)) {
+		if (existsat(rc_dirfd(rc_service_state_names[i].dir), base)) {
 			if (rc_service_state_names[i].state <= 0x10)
 				state = rc_service_state_names[i].state;
 			else
@@ -851,22 +1005,23 @@ rc_service_state(const char *service)
 		}
 	}
 
-	if (state & RC_SERVICE_STARTED) {
+	if (state & RC_SERVICE_STARTED)
 		if (rc_service_daemons_crashed(service) && errno != EACCES)
 			state |= RC_SERVICE_CRASHED;
-	}
+
 	if (state & RC_SERVICE_STOPPED) {
-		dirs = ls_dir(RC_SVCDIR "/scheduled", 0);
-		TAILQ_FOREACH(dir, dirs, entries) {
-			snprintf(file, sizeof(file),
-			    RC_SVCDIR "/scheduled/%s/%s",
-			    dir->value, service);
-			if (exists(file)) {
-				state |= RC_SERVICE_SCHEDULED;
-				break;
+		DIR *dp = do_dopendir(rc_dirfd(RC_DIR_SCHEDULED));
+		if (dp) {
+			struct dirent *d;
+			while ((d = readdir(dp))) {
+				if (existsat(dirfd(dp), service)) {
+					state |= RC_SERVICE_SCHEDULED;
+					break;
+				}
 			}
+
+			closedir(dp);
 		}
-		rc_stringlist_free(dirs);
 	}
 
 	return state;
@@ -877,61 +1032,65 @@ rc_service_value_get(const char *service, const char *option)
 {
 	char *buffer = NULL;
 	size_t len = 0;
-	char file[PATH_MAX];
+	char *file;
 
-	snprintf(file, sizeof(file), RC_SVCDIR "/options/%s/%s",
-	    service, option);
+	xasprintf(&file, "%s/options/%s/%s", rc_svcdir(), service, option);
 	rc_getfile(file, &buffer, &len);
+	free(file);
 
 	return buffer;
 }
 
 bool
-rc_service_value_set(const char *service, const char *option,
-    const char *value)
+rc_service_value_set(const char *service, const char *option, const char *value)
 {
+	bool ret = true;
+	int optdirfd;
 	FILE *fp;
-	char file[PATH_MAX];
-	char *p = file;
 
-	p += snprintf(file, sizeof(file), RC_SVCDIR "/options/%s", service);
-	if (mkdir(file, 0755) != 0 && errno != EEXIST)
+	if (mkdirat(rc_dirfd(RC_DIR_OPTIONS), service, 0755) != 0 && errno != EEXIST)
 		return false;
 
-	snprintf(p, sizeof(file) - (p - file), "/%s", option);
-	if (value) {
-		if (!(fp = fopen(file, "w")))
-			return false;
+	if ((optdirfd = openat(rc_dirfd(RC_DIR_OPTIONS), service, O_RDONLY | O_DIRECTORY)) == -1)
+		return false;
+
+	if (!value) {
+		unlinkat(optdirfd, option, 0);
+	} else if ((fp = do_fopenat(optdirfd, option, O_WRONLY | O_CREAT | O_TRUNC))) {
 		fprintf(fp, "%s", value);
 		fclose(fp);
 	} else {
-		unlink(file);
+		ret = false;
 	}
-		return true;
-}
 
+	close(optdirfd);
+	return ret;
+}
 
 bool
 rc_service_schedule_start(const char *service, const char *service_to_start)
 {
-	char file[PATH_MAX];
-	char *p = file;
-	char *init;
+	const char *base_to_start = basename_c(service_to_start);
+	const char *base = basename_c(service);
 	bool retval;
+	int schedfd;
+	char *init;
 
 	/* service may be a provided service, like net */
 	if (!service || !rc_service_exists(service_to_start))
 		return false;
 
-	p += snprintf(file, sizeof(file), RC_SVCDIR "/scheduled/%s",
-	    basename_c(service));
-	if (mkdir(file, 0755) != 0 && errno != EEXIST)
+	if (mkdirat(rc_dirfd(RC_DIR_SCHEDULED), base, 0755) != 0 && errno != EEXIST)
+		return false;
+
+	if ((schedfd = openat(rc_dirfd(RC_DIR_SCHEDULED), base, O_RDONLY)) == -1)
 		return false;
 
 	init = rc_service_resolve(service_to_start);
-	snprintf(p, sizeof(file) - (p - file),
-	    "/%s", basename_c(service_to_start));
-	retval = (exists(file) || symlink(init, file) == 0);
+
+	retval = existsat(schedfd, base_to_start) == 0 || symlinkat(init, schedfd, base_to_start) == 0;
+
+	close(schedfd);
 	free(init);
 	return retval;
 }
@@ -939,50 +1098,32 @@ rc_service_schedule_start(const char *service, const char *service_to_start)
 bool
 rc_service_schedule_clear(const char *service)
 {
-	char dir[PATH_MAX];
-
-	snprintf(dir, sizeof(dir), RC_SVCDIR "/scheduled/%s",
-	    basename_c(service));
-	if (!rm_dir(dir, true) && errno == ENOENT)
-		return true;
-	return false;
+	return !rm_dir(rc_dirfd(RC_DIR_SCHEDULED), basename_c(service), true) && errno == ENOENT;
 }
 
 RC_STRINGLIST *
 rc_services_in_runlevel(const char *runlevel)
 {
-	char dir[PATH_MAX];
 	RC_STRINGLIST *list = NULL;
 
 	if (!runlevel) {
-#ifdef RC_PKG_INITDIR
-		RC_STRINGLIST *pkg = ls_dir(RC_PKG_INITDIR, LS_INITD);
-#endif
-#ifdef RC_LOCAL_INITDIR
-		RC_STRINGLIST *local = ls_dir(RC_LOCAL_INITDIR, LS_INITD);
-#endif
+		const int *scriptdirfds;
+		list = rc_stringlist_new();
+		for (size_t i = 0, count = rc_scriptdirfds(&scriptdirfds); i < count; i++) {
+			RC_STRINGLIST *svcs;
 
-		list = ls_dir(RC_INITDIR, LS_INITD);
-
-#ifdef RC_PKG_INITDIR
-		TAILQ_CONCAT(list, pkg, entries);
-		rc_stringlist_free(pkg);
-#endif
-#ifdef RC_LOCAL_INITDIR
-		TAILQ_CONCAT(list, local, entries);
-		rc_stringlist_free(local);
-#endif
+			svcs = ls_dir(scriptdirfds[i], "init.d", LS_INITD);
+			TAILQ_CONCAT(list, svcs, entries);
+			rc_stringlist_free(svcs);
+		}
 		return list;
 	}
 
 	/* These special levels never contain any services */
-	if (strcmp(runlevel, RC_LEVEL_SINGLE) != 0) {
-		snprintf(dir, sizeof(dir), RC_RUNLEVELDIR "/%s", runlevel);
-		list = ls_dir(dir, LS_INITD);
-	}
-	if (!list)
-		list = rc_stringlist_new();
-	return list;
+	if (strcmp(runlevel, RC_LEVEL_SINGLE) == 0)
+		return rc_stringlist_new();
+
+	return ls_dir(rc_dirfd(RC_DIR_RUNLEVEL), runlevel, LS_INITD);
 }
 
 RC_STRINGLIST *
@@ -1007,31 +1148,24 @@ rc_services_in_state(RC_SERVICE state)
 {
 	RC_STRINGLIST *services;
 	RC_STRINGLIST *list;
-	RC_STRINGLIST *dirs;
-	RC_STRING *d;
-	char dir[PATH_MAX];
-	char *p = dir;
-
-	p += snprintf(dir, sizeof(dir), RC_SVCDIR "/%s",
-	    rc_parse_service_state(state));
+	struct dirent *d;
+	DIR *dp;
 
 	if (state != RC_SERVICE_SCHEDULED)
-		return ls_dir(dir, LS_INITD);
+		return ls_dir(rc_dirfd(RC_DIR_SVCDIR), rc_parse_service_state(state), LS_INITD);
 
-	dirs = ls_dir(dir, 0);
 	list = rc_stringlist_new();
-	if (!dirs)
+	if (!(dp = do_opendirat(rc_dirfd(RC_DIR_SVCDIR), "scheduled")))
 		return list;
 
-	TAILQ_FOREACH(d, dirs, entries) {
-		snprintf(p, sizeof(dir) - (p - dir), "/%s", d->value);
-		services = ls_dir(dir, LS_INITD);
-		if (services) {
-			TAILQ_CONCAT(list, services, entries);
-			rc_stringlist_free(services);
-		}
+	while ((d = readdir(dp))) {
+		if (!(services = ls_dir(dirfd(dp), d->d_name, LS_INITD)))
+			continue;
+		TAILQ_CONCAT(list, services, entries);
+		rc_stringlist_free(services);
 	}
-	rc_stringlist_free(dirs);
+	closedir(dp);
+
 	return list;
 }
 
@@ -1040,9 +1174,9 @@ rc_service_add(const char *runlevel, const char *service)
 {
 	bool retval;
 	char *init;
-	char file[PATH_MAX];
-	char path[MAXPATHLEN] = { '\0' };
-	char binit[PATH_MAX];
+	char *file;
+	char *path;
+	char *binit = NULL;
 	char *i;
 
 	if (!rc_runlevel_exists(runlevel)) {
@@ -1056,26 +1190,39 @@ rc_service_add(const char *runlevel, const char *service)
 	}
 
 	i = init = rc_service_resolve(service);
-	snprintf(file, sizeof(file), RC_RUNLEVELDIR "/%s/%s",
-	    runlevel, basename_c(service));
 
 	/* We need to ensure that only things in /etc/init.d are added
 	 * to the boot runlevel */
 	if (strcmp(runlevel, RC_LEVEL_BOOT) == 0) {
-		if (realpath(dirname(init), path) == NULL) {
+		path = realpath(dirname(init), NULL);
+		if (path == NULL) {
 			free(init);
 			return false;
 		}
-		if (strcmp(path, RC_INITDIR) != 0) {
+		retval = (strcmp(path, RC_INITDIR) == 0);
+		free(path);
+		if (!retval) {
 			free(init);
 			errno = EPERM;
 			return false;
 		}
-		snprintf(binit, sizeof(binit), RC_INITDIR "/%s", service);
+		xasprintf(&binit, "%s/%s", RC_INITDIR, service);
 		i = binit;
 	}
 
+	/* We shouldn't add any temporary script to a runlevel */
+	if (strncmp(i, rc_svcdir(), strlen(rc_svcdir())) == 0) {
+		errno = ENOENT;
+		retval = false;
+		goto out;
+	}
+
+	xasprintf(&file, "%s/%s/%s", rc_runleveldir(), runlevel, basename_c(service));
 	retval = (symlink(i, file) == 0);
+	free(file);
+
+out:
+	free(binit);
 	free(init);
 	return retval;
 }
@@ -1083,39 +1230,39 @@ rc_service_add(const char *runlevel, const char *service)
 bool
 rc_service_delete(const char *runlevel, const char *service)
 {
-	char file[PATH_MAX];
+	char *file;
+	int r;
 
-	snprintf(file, sizeof(file), RC_RUNLEVELDIR "/%s/%s",
-	    runlevel, basename_c(service));
-	if (unlink(file) == 0)
-		return true;
-	return false;
+	xasprintf(&file, "%s/%s/%s", rc_runleveldir(), runlevel,
+		basename_c(service));
+	r = unlink(file);
+	free(file);
+	return (r == 0);
 }
 
 RC_STRINGLIST *
 rc_services_scheduled_by(const char *service)
 {
-	RC_STRINGLIST *dirs = ls_dir(RC_SVCDIR "/scheduled", 0);
 	RC_STRINGLIST *list = rc_stringlist_new();
-	RC_STRING *dir;
-	char file[PATH_MAX];
+	struct dirent *d;
+	DIR *dp;
 
-	TAILQ_FOREACH(dir, dirs, entries) {
-		snprintf(file, sizeof(file), RC_SVCDIR "/scheduled/%s/%s",
-		    dir->value, service);
-		if (exists(file))
-			rc_stringlist_add(list, file);
+	if (!(dp = do_dopendir(rc_dirfd(RC_DIR_SCHEDULED))))
+		return list;
+
+	while ((d = readdir(dp))) {
+		char *file;
+		xasprintf(&file, "%s/%s", d->d_name, service);
+		if (existsat(dirfd(dp), file))
+			rc_stringlist_add(list, d->d_name);
+		free(file);
 	}
-	rc_stringlist_free(dirs);
+
 	return list;
 }
 
 RC_STRINGLIST *
 rc_services_scheduled(const char *service)
 {
-	char dir[PATH_MAX];
-
-	snprintf(dir, sizeof(dir), RC_SVCDIR "/scheduled/%s",
-	    basename_c(service));
-	return ls_dir(dir, LS_INITD);
+	return ls_dir(rc_dirfd(RC_DIR_SCHEDULED), basename_c(service), LS_INITD);
 }
