@@ -19,6 +19,9 @@
 #include <errno.h>
 #include <fcntl.h>
 #include <limits.h>
+#ifdef HAVE_LINUX_CLOSE_RANGE_H
+#  include <linux/close_range.h>
+#endif
 #include <signal.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -26,10 +29,13 @@
 #include <sys/file.h>
 #include <sys/time.h>
 #ifdef __linux__
+#  include <sys/syscall.h> /* for close_range */
 #  include <sys/sysinfo.h>
 #endif
 #include <sys/types.h>
 #include <sys/utsname.h>
+#include <sys/un.h>
+#include <sys/socket.h>
 #include <time.h>
 #include <unistd.h>
 #include <utime.h>
@@ -49,21 +55,37 @@ rc_conf_yesno(const char *setting)
 	return rc_yesno(rc_conf_value (setting));
 }
 
-static const char *const env_whitelist[] = {
+static const char *const env_allowlist[] = {
 	"EERROR_QUIET", "EINFO_QUIET",
 	"IN_BACKGROUND", "IN_DRYRUN", "IN_HOTPLUG",
-	"RC_DEBUG", "RC_NODEPS",
+	"RC_DEBUG", "RC_NODEPS", "RC_USER_SERVICES",
 	"LANG", "LC_MESSAGES", "TERM",
 	"EINFO_COLOR", "EINFO_VERBOSE",
 	NULL
 };
 
+static const char *const usrenv_allowlist[] = {
+	"USER", "LOGNAME", "HOME", "SHELL", "XDG_RUNTIME_DIR",
+	"XDG_CONFIG_HOME", "XDG_STATE_HOME", "XDG_CACHE_HOME",
+	NULL
+};
+
+static bool
+env_allowed(const char *const list[], const char *value)
+{
+	for (size_t i = 0; list[i]; i++)
+		if (strcmp(list[i], value) == 0)
+			return true;
+	return false;
+}
+
 void
 env_filter(void)
 {
 	RC_STRINGLIST *env_allow;
 	RC_STRINGLIST *profile;
 	RC_STRINGLIST *env_list;
+	char *profile_path;
 	RC_STRING *env;
 	char *e;
 	size_t i = 0;
@@ -78,7 +100,19 @@ env_filter(void)
 		rc_stringlist_free(env_allow);
 		return;
 	}
-	profile = rc_config_load(RC_PROFILE_ENV);
+
+	xasprintf(&profile_path, "%s/profile.env", rc_sysconfdir());
+	profile = rc_config_load(profile_path);
+	free(profile_path);
+
+	if (rc_is_user()) {
+		RC_STRINGLIST *usrprofile;
+		xasprintf(&profile_path, "%s/profile.env", rc_usrconfdir());
+		usrprofile = rc_config_load(profile_path);
+		free(profile_path);
+		TAILQ_CONCAT(profile, usrprofile, entries);
+		rc_stringlist_free(usrprofile);
+	}
 
 	/* Copy the env and work from this so we can manipulate it safely */
 	env_list = rc_stringlist_new();
@@ -91,11 +125,9 @@ env_filter(void)
 
 	TAILQ_FOREACH(env, env_list, entries) {
 		/* Check the whitelist */
-		for (i = 0; env_whitelist[i]; i++) {
-			if (strcmp(env_whitelist[i], env->value) == 0)
-				break;
-		}
-		if (env_whitelist[i])
+		if (env_allowed(env_allowlist, env->value))
+			continue;
+		if (rc_is_user() && env_allowed(usrenv_allowlist, env->value))
 			continue;
 
 		/* Check our user defined list */
@@ -134,7 +166,9 @@ env_config(void)
 	char *npp;
 	char *tok;
 	const char *sys = rc_sys();
+	const char *svcdir = rc_svcdir();
 	char *buffer = NULL;
+	char *tmpdir;
 	size_t size = 0;
 
 	/* Ensure our PATH is prefixed with the system locations first
@@ -164,34 +198,48 @@ env_config(void)
 		free(e);
 	}
 
+	if (!rc_is_user()) {
+		setenv("RC_CACHEDIR", "/var/cache/rc", 1);
+	} else {
+		const char *cache_home = getenv("XDG_CACHE_HOME");
+		char *cachedir = NULL;
+
+		if (cache_home)
+			xasprintf(&cachedir, "%s/rc", cache_home);
+		else if ((cache_home = getenv("HOME")))
+			xasprintf(&cachedir, "%s/.cache/rc", cache_home);
+
+		if (cachedir)
+			setenv("RC_CACHEDIR", cachedir, 1);
+
+		free(cachedir);
+	}
+
+	xasprintf(&tmpdir, "%s/tmp", svcdir);
+	e = rc_runlevel_get();
+
 	setenv("RC_VERSION", VERSION, 1);
 	setenv("RC_LIBEXECDIR", RC_LIBEXECDIR, 1);
-	setenv("RC_SVCDIR", RC_SVCDIR, 1);
-	setenv("RC_TMPDIR", RC_SVCDIR "/tmp", 1);
+	setenv("RC_SVCDIR", svcdir, 1);
+	setenv("RC_TMPDIR", tmpdir, 1);
 	setenv("RC_BOOTLEVEL", RC_LEVEL_BOOT, 1);
-	e = rc_runlevel_get();
 	setenv("RC_RUNLEVEL", e, 1);
+
 	free(e);
+	free(tmpdir);
 
 	if ((fp = fopen(RC_KRUNLEVEL, "r"))) {
-		if (getline(&buffer, &size, fp) != -1) {
-			l = strlen (buffer) - 1;
-			if (buffer[l] == '\n')
-				buffer[l] = 0;
+		if (xgetline(&buffer, &size, fp) != -1)
 			setenv("RC_DEFAULTLEVEL", buffer, 1);
-		}
+		free(buffer);
 		fclose(fp);
-	} else
+	} else {
 		setenv("RC_DEFAULTLEVEL", RC_LEVEL_DEFAULT, 1);
+	}
 
-	free(buffer);
 	if (sys)
 		setenv("RC_SYS", sys, 1);
 
-#ifdef PREFIX
-	setenv("RC_PREFIX", RC_PREFIX, 1);
-#endif
-
 	/* Some scripts may need to take a different code path if
 	   Linux/FreeBSD, etc
 	   To save on calling uname, we store it in an environment variable */
@@ -221,27 +269,11 @@ signal_setup(int sig, void (*handler)(int))
 	return sigaction(sig, &sa, NULL);
 }
 
-int
-signal_setup_restart(int sig, void (*handler)(int))
-{
-	struct sigaction sa;
-
-	memset(&sa, 0, sizeof (sa));
-	sigemptyset(&sa.sa_mask);
-	sa.sa_handler = handler;
-	sa.sa_flags = SA_RESTART;
-	return sigaction(sig, &sa, NULL);
-}
-
 int
 svc_lock(const char *applet, bool ignore_lock_failure)
 {
-	char *file = NULL;
-	int fd;
+	int fd = openat(rc_dirfd(RC_DIR_EXCLUSIVE), applet, O_WRONLY | O_CREAT | O_NONBLOCK, 0664);
 
-	xasprintf(&file, RC_SVCDIR "/exclusive/%s", applet);
-	fd = open(file, O_WRONLY | O_CREAT | O_NONBLOCK, 0664);
-	free(file);
 	if (fd == -1)
 		return -1;
 	if (flock(fd, LOCK_EX | LOCK_NB) == -1) {
@@ -263,12 +295,8 @@ svc_lock(const char *applet, bool ignore_lock_failure)
 int
 svc_unlock(const char *applet, int fd)
 {
-	char *file = NULL;
-
-	xasprintf(&file, RC_SVCDIR "/exclusive/%s", applet);
+	unlinkat(rc_dirfd(RC_DIR_EXCLUSIVE), applet, 0);
 	close(fd);
-	unlink(file);
-	free(file);
 	return -1;
 }
 
@@ -312,8 +340,11 @@ exec_service(const char *service, const char *arg)
 		sigaction(SIGUSR1, &sa, NULL);
 		sigaction(SIGWINCH, &sa, NULL);
 
-		/* Unmask signals */
-		sigprocmask(SIG_SETMASK, &old, NULL);
+		/* Unmask all signals.
+		 * We might've been called from pam_openrc by
+		 * a process that masked signals we rely on.
+		 * Bug: https://bugs.gentoo.org/953748 */
+		sigprocmask(SIG_UNBLOCK, &full, NULL);
 
 		/* Safe to run now */
 		execl(file, file, "--lockfd", sfd, arg, (char *) NULL);
@@ -327,7 +358,7 @@ exec_service(const char *service, const char *arg)
 		fprintf(stderr, "fork: %s\n",strerror (errno));
 		svc_unlock(basename_c(service), fd);
 	} else
-		fcntl(fd, F_SETFD, fcntl(fd, F_GETFD, 0) | FD_CLOEXEC);
+		close(fd);
 
 	sigprocmask(SIG_SETMASK, &old, NULL);
 	free(file);
@@ -373,14 +404,14 @@ RC_DEPTREE * _rc_deptree_load(int force, int *regen)
 	int merrno;
 	time_t t;
 	char file[PATH_MAX];
+	int svcdirfd = rc_dirfd(RC_DIR_SVCDIR);
 	struct stat st;
-	struct utimbuf ut;
 	FILE *fp;
 
 	t = 0;
 	if (rc_deptree_update_needed(&t, file) || force != 0) {
 		/* Test if we have permission to update the deptree */
-		fd = open(RC_DEPTREE_CACHE, O_WRONLY);
+		fd = openat(svcdirfd, "deptree", O_WRONLY);
 		merrno = errno;
 		errno = serrno;
 		if (fd == -1 && merrno == EACCES)
@@ -391,33 +422,33 @@ RC_DEPTREE * _rc_deptree_load(int force, int *regen)
 			*regen = 1;
 		ebegin("Caching service dependencies");
 		retval = rc_deptree_update() ? 0 : -1;
-		eend (retval, "Failed to update the dependency tree");
+		eend(retval, "Failed to update the dependency tree");
 
 		if (retval == 0) {
-			if (stat(RC_DEPTREE_CACHE, &st) != 0) {
-				eerror("stat(%s): %s", RC_DEPTREE_CACHE, strerror(errno));
+			if (fstatat(svcdirfd, "deptree", &st, 0) != 0) {
+				eerror("stat(%s): %s/deptree", rc_svcdir(), strerror(errno));
 				return NULL;
 			}
-			if (st.st_mtime < t) {
-				eerror("Clock skew detected with `%s'", file);
-				eerrorn("Adjusting mtime of `" RC_DEPTREE_CACHE
-				    "' to %s", ctime(&t));
-				fp = fopen(RC_DEPTREE_SKEWED, "w");
-				if (fp != NULL) {
-					fprintf(fp, "%s\n", file);
-					fclose(fp);
-				}
-				ut.actime = t;
-				ut.modtime = t;
-				utime(RC_DEPTREE_CACHE, &ut);
-			} else {
-				if (exists(RC_DEPTREE_SKEWED))
-					unlink(RC_DEPTREE_SKEWED);
+
+			if (st.st_mtime >= t) {
+				unlinkat(svcdirfd, "clock-skewed", 0);
+				goto out;
+			}
+
+			eerror("Clock skew detected with '%s/clock-skewed'", rc_svcdir());
+			eerrorn("Adjusting mtime of '%s/deptree' to %s", rc_svcdir(), ctime(&t));
+			if ((fp = do_fopenat(svcdirfd, "clock-skewed", O_WRONLY | O_CREAT | O_TRUNC))) {
+				fprintf(fp, "%s\n", file);
+				futimens(fileno(fp), (struct timespec[]) {{ .tv_sec = t }, { .tv_sec = t }});
+				fclose(fp);
 			}
 		}
+
+out:
 		if (force == -1 && regen != NULL)
 			*regen = retval;
 	}
+
 	return rc_deptree_load();
 }
 
@@ -500,3 +531,114 @@ pid_t get_pid(const char *applet,const char *pidfile)
 
 	return pid;
 }
+
+struct notify notify_parse(const char *applet, const char *notify_string)
+{
+	struct notify notify = {0};
+	if (sscanf(notify_string, "fd:%d", &notify.fd) == 1) {
+		notify.type = NOTIFY_FD;
+		if (pipe(notify.pipe) == -1)
+			eerrorx("%s: pipe: %s", applet, strerror(errno));
+	} else if (strncmp(notify_string, "socket", sizeof("socket") - 1) == 0) {
+		union {
+			struct sockaddr header;
+			struct sockaddr_un unix;
+		} addr = { .unix = { .sun_family = AF_UNIX } };
+		int written;
+		const char *opts = strchr(notify_string, ':');
+
+		if (!opts || opts[1] == '\0')
+			return notify;
+		if (strcmp(opts + 1, "ready") != 0)
+			return notify;
+
+		written = snprintf(addr.unix.sun_path, sizeof(addr.unix.sun_path), "%s/supervise-%s.sock", rc_svcdir(), applet);
+		if (written >= (int)sizeof(addr.unix.sun_path))
+			eerrorx("%s: socket path '%s/supervise-%s.sock' too long.", applet, rc_svcdir(), applet);
+		setenv("NOTIFY_SOCKET", addr.unix.sun_path, true);
+
+		notify.type = NOTIFY_SOCKET;
+		if ((notify.fd = socket(AF_UNIX, SOCK_DGRAM, 0)) == -1)
+			eerrorx("%s: socket: %s", applet, strerror(errno));
+		if (bind(notify.fd, &addr.header, sizeof(addr.unix)) == -1)
+			eerrorx("%s: bind: %s", applet, strerror(errno));
+	} else {
+		eerrorx("%s: invalid ready '%s'.", applet, optarg);
+	}
+
+	return notify;
+}
+
+bool notify_wait(const char *applet, struct notify notify)
+{
+	switch (notify.type) {
+	case NOTIFY_NONE:
+		return true;
+	case NOTIFY_FD:
+		close(notify.pipe[1]);
+		notify.fd = notify.pipe[0];
+		break;
+	case NOTIFY_SOCKET:
+		break;
+	}
+
+	for (;;) {
+		char buf[BUFSIZ];
+		ssize_t bytes = read(notify.fd, buf, BUFSIZ);
+		if (bytes == -1) {
+			if (errno != EINTR) {
+				eerror("%s: read failed '%s'\n", applet, strerror(errno));
+				return false;
+			}
+			continue;
+		}
+
+		switch (notify.type) {
+		case NOTIFY_NONE:
+			break;
+		case NOTIFY_FD:
+			if (memchr(buf, '\n', bytes))
+				return true;
+			break;
+		case NOTIFY_SOCKET:
+			buf[bytes] = '\0';
+			if (strstr(buf, "READY=1")) {
+				char *path;
+				xasprintf(&path, "%s/supervise-%s.sock", rc_svcdir(), applet);
+				unlink(path);
+				free(path);
+				return true;
+			}
+			break;
+		}
+	}
+
+	return true;
+}
+
+#ifndef HAVE_CLOSE_RANGE
+static inline int close_range(unsigned int first RC_UNUSED,
+			      unsigned int last RC_UNUSED,
+			      unsigned int flags RC_UNUSED)
+{
+#ifdef SYS_close_range
+	return syscall(SYS_close_range, first, last, flags);
+#else
+	errno = ENOSYS;
+	return -1;
+#endif
+}
+#endif
+#ifndef CLOSE_RANGE_CLOEXEC
+# define CLOSE_RANGE_CLOEXEC	(1U << 2)
+#endif
+
+void
+cloexec_fds_from(int first)
+{
+	int i;
+	if (close_range(first, UINT_MAX, CLOSE_RANGE_CLOEXEC) < 0) {
+		for (i = getdtablesize() - 1; i >= first; --i)
+			fcntl(i, F_SETFD, FD_CLOEXEC);
+	}
+}
